#include "rules.h"

extern PrintIP pip;

ProgRules pr;
RulePtr current;
int file_line;
Rule *rl;
int rl_flag;

/****************************************************************************
 *
 * Function: ParseRulesFile(char *)
 *
 * Purpose:  Read the rules file a line at a time and send each rule to
 *           the rule parser
 *
 * Arguments: file => rules file filename
 *
 * Returns: void function
 *
 ****************************************************************************/
void ParseRulesFile(char *file)
{
   FILE *thefp;
   char buf[STD_BUF];

   printf("Opening rules file: %s\n", file);

   if((thefp = fopen(file,"r")) == NULL)
   {
      printf("Unable to open rules file: %s\n", file);
      exit(1);
   }

   bzero(buf, STD_BUF);

   while((fgets(buf, STD_BUF, thefp)) != NULL)
   {
      file_line++;

#ifdef DEBUG2
      printf("Got line %d: %s", file_line, buf);
#endif
      if((buf[0] != '#') && (buf[0] != 0x0a) && (buf[0] != ';'))
      {
         ParseRule(buf);
      }

      bzero(buf, STD_BUF);
   }

#ifdef DEBUG
   DumpRuleList();
#endif   

   fclose(thefp);

   return;
}



/****************************************************************************
 *
 * Function: ParseRule(char *)
 *
 * Purpose:  Process an individual rule and add it to the rule list
 *
 * Arguments: rule => rule string
 *
 * Returns: void function
 *
 ****************************************************************************/
void ParseRule(char *rule)
{
   char **toks;
   int num_toks;
   char *alert_ptr;
   char *alert_end;
   u_int alert_str_size;


   /* chop off the <CR/LF> from the string */
   strip(rule);

   /* break out the tokens from the rule string */
   toks = mSplit(rule, " ", 10, &num_toks);

   /* Make a rule node */
   CreateRuleNode();
                       
   /* figure out what type (pass/log/alert) rule is being looked at */
   current->type = RuleType(toks[0]);
 
   /* set the rule protocol */
   current->proto = WhichProto(toks[1]);

   /* Process the IP address and CIDER netmask */
   /* if this is an "any" addr, set the flag */
   if(ParseIP(toks[2], (u_long *) &current->sip, (u_long *) &current->smask))
      current->flags |= ANY_SRC_IP;

   /* do the same for the port */
   if(ParsePort(toks[3], (u_short *) &current->sp))
      current->flags |= ANY_SRC_PORT;

   if(ParseIP(toks[5], (u_long *) &current->dip, (u_long *) &current->dmask))
      current->flags |= ANY_DST_IP;

   if(ParsePort(toks[6], (u_short *) &current->dp))
      current->flags |= ANY_DST_PORT;

   /* grab the alert message */
   if(current->type == RULE_ALERT)
   {
      alert_ptr = index(rule, '<');
      alert_ptr++;
      alert_end = strrchr(rule, '>');
      alert_str_size = (alert_end - alert_ptr);

      if(alert_str_size > 0)
      {
         current->message = (char *)malloc((sizeof(char)*alert_str_size));
         strncpy(current->message, alert_ptr, alert_str_size);
      }
      else 
      {
         fprintf(stderr, "ERROR Line %d: bad alert message size %d\n", file_line, alert_str_size);
         exit(1);
      }
   }

   free(toks);

   return;
}


/****************************************************************************
 *
 * Function: RuleType(char *)
 *
 * Purpose:  Determine what type of rule is being processed and return its
 *           equivalent value
 *
 * Arguments: func => string containing the rule type
 *
 * Returns: The rule type designation
 *
 ****************************************************************************/
int RuleType(char *func)
{
   if(!strncasecmp(func, "log",3))
      return RULE_LOG;

   if(!strncasecmp(func, "alert",5))
      return RULE_ALERT;

   if(!strncasecmp(func, "pass",4))
      return RULE_PASS;

   
   printf("ERROR line %d => Unknown Rule action: %s\n", file_line, func);
   CleanExit();
  
   return 0;
}

      

/****************************************************************************
 *
 * Function: CreateRuleNode()
 *
 * Purpose:  Allocate space for a rule and attach it to the end of the
 *           program rule list
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 ****************************************************************************/
void CreateRuleNode()
{
   RulePtr idx;

   current = (RulePtr) malloc(sizeof(Rule));

   if(current == NULL)
   {
      perror("CreateRuleNode()");
      exit(1);
   }

   current->next = NULL;

   if(!rl_flag)
   {
      rl_flag = 1;

      rl = (Rule *) malloc(sizeof(Rule));

      memcpy(rl, current, sizeof(Rule));

      free(current);

      current = rl;
 
      return;
   }      

   idx = rl;

   if(idx == 0x0)
   {
      idx = current;
      return;
   }
   else
   {
      while(idx->next != NULL)
      {
         idx = idx->next;
      }

      idx->next = current;
   }

   return;
}



/****************************************************************************
 *
 * Function: WhichProto(char *)
 *
 * Purpose: Figure out which protocol the current rule is talking about
 *
 * Arguments: proto_str => the protocol string
 *
 * Returns: The integer value of the protocol
 *
 ****************************************************************************/
int WhichProto(char *proto_str)
{
   if(!strncasecmp(proto_str, "tcp", 3))
      return IPPROTO_TCP;

   if(!strncasecmp(proto_str, "udp", 3))
      return IPPROTO_UDP;

   if(!strncasecmp(proto_str, "icmp", 4))
      return IPPROTO_ICMP;

   fprintf(stderr, "ERROR Line %d => Bad protocol: %s\n", file_line, proto_str);
   exit(1);
}


/****************************************************************************
 *
 * Function: ParseIP(char *, u_long *, u_long *)
 *
 * Purpose: Convert a supplied IP address to it's network order 32-bit long
 *          value.  Also convert the CIDER block notation into a real 
 *          netmask. 
 *
 * Arguments: addr => address string to convert
 *            ip_addr => storage point for the converted ip address
 *            netmask => storage point for the converted netmask
 *
 * Returns: 0 for normal addresses, 1 for an "any" address
 *
 ****************************************************************************/
int ParseIP(char *addr, u_long *ip_addr, u_long *netmask)
{
   char **toks;
   int num_toks;
   int nmask;

   if(!strncasecmp(addr, "any", 3))
   {
      *ip_addr = 0;
      *netmask = 0;
      return 1;
   }
 
   /* break out the CIDER notation from the IP address */
   toks = mSplit(addr,"/",2,&num_toks);

   /* convert the CIDER notation into a real live netmask */
   nmask = 32 - atoi(toks[1]);

   *netmask = pow(2, nmask) - 1;
   *netmask = -(*netmask);
   *netmask -= 1;

#ifndef WORDS_BIGENDIAN
   /* since PC's store things the "wrong" way, shuffle the bytes into
      the right order */
   *netmask = htonl(*netmask);
#endif

   /* convert the IP addr into its 32-bit value */
   if((*ip_addr = inet_addr(toks[0])) == -1)
   {
      fprintf(stderr, "ERROR Line %d => Rule IP addr (%s) didn't x-late, WTF?\n",
              file_line, toks[0]);
      exit(0);
   }
   else
   {
      /* set the final homenet address up */
      *ip_addr = ((u_long)(*ip_addr) & (*netmask));
   }

   free(toks);

   return 0;
}



/****************************************************************************
 *
 * Function: ParsePort(char *, u_short *)
 *
 * Purpose:  Convert the port string over to an integer value
 *
 * Arguments: rule_port => port string
 *            port => converted integer value of the port
 *
 * Returns: 0 for a normal port number, 1 for an "any" port
 *
 ****************************************************************************/
int ParsePort(char *rule_port, u_short *port)
{
   int conv;

   if(!strncasecmp(rule_port, "any", 3))
   {
      *port = 0;
      return 1;
   }

   conv = atoi(rule_port);

   if((conv >= 0) && (conv < 65536))
   {
      *port = conv;
   }
   else
   {
      fprintf(stderr, "ERROR Line %d => bad port number: %s", file_line, rule_port);
      exit(1);
   }

   return 0;
}



/****************************************************************************
 *
 * Function: CheckRules()
 *
 * Purpose:  Test the current packet against the rule list to classify it.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 ****************************************************************************/
void CheckRules()
{
   Rule *idx;

   idx = rl;

   while(idx != NULL)
   {
      if(idx->proto == net.proto)
      {
#ifdef DEBUG
         printf("Proto match\n");
#endif
         if((idx->flags & ANY_SRC_IP) || 
            (idx->sip == (net.sip & idx->smask)))
         {
#ifdef DEBUG
            printf("SIP match\n");
#endif
            if((idx->flags & ANY_SRC_PORT) ||
               (idx->sp == net.sp))
            {
#ifdef DEBUG
               printf("SP match\n");
#endif
               if((idx->flags & ANY_DST_IP) ||
                  (idx->dip == (net.dip & idx->dmask)))
               {
#ifdef DEBUG
                  printf("DIP match\n");
#endif
                  if((idx->flags & ANY_DST_PORT) ||
                     (idx->dp == net.dp))
                  {
#ifdef DEBUG
                     printf("DP match\n");
#endif
                     current = idx;
                     switch(idx->type)
                     {
                        case RULE_PASS:  return;
                      
                        case RULE_ALERT: AlertPkt();
                                         return;

                        case RULE_LOG:  LogPkt();
                                        return;
                      }
                   }  
#ifdef DEBUG
                   else
                      printf("DP mismatch!\n");
#endif
                }
#ifdef DEBUG
                else
                   printf("DIP mismatch!\n");
#endif
             }
#ifdef DEBUG
             else
                printf("SP mismatch!\n");
#endif
          }
#ifdef DEBUG
          else
             printf("SIP mismatch\n");
#endif
      }   
#ifdef DEBUG
      else
         printf("proto mismatch!\n");
#endif

      idx = idx->next;
   } 

   return;
}   



/****************************************************************************
 *
 * Function: LogPkt()
 *
 * Purpose: Log the current packet to the disk
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 ****************************************************************************/
void LogPkt()
{
   OpenLogFile();

   PrintIPPkt(log_ptr, net.proto);

   if(pv.data_flag)
   {
      PrintNetData(log_ptr, pip.data, pip.dsize);
      printf("\n");
   }

   fclose(log_ptr);
}



/****************************************************************************
 *
 * Function: AlertPkt()
 *
 * Purpose: Send the current packet to the alert file
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 ****************************************************************************/
void AlertPkt()
{
   OpenAlertFile();

   fprintf(alert, "[**][%s][**]\n", current->message);

   PrintIPPkt(alert, net.proto);

   if(pv.data_flag)
   {
      PrintNetData(alert, pip.data, pip.dsize);
      printf("\n");
   }

   fclose(alert);
}
      


/****************************************************************************
 *
 * Function: DumpRuleList()
 *
 * Purpose: Debug function, displays the rule list
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 ****************************************************************************/
void DumpRuleList()
{
   Rule *idx;
   struct in_addr temp;

   idx = rl;

   while(idx != NULL)
   {
      printf("RULE: type=%d\n",idx->type);
      printf("      proto=%d\n", idx->proto);
      temp.s_addr = idx->sip;
      printf("      %s:%d ",inet_ntoa(temp),idx->sp);
      temp.s_addr = idx->dip;
      printf("-> %s:%d\n",inet_ntoa(temp), idx->dp);
      printf("      flags= ");
      if(idx->flags & ANY_SRC_IP)
         printf("ANY_SRC_IP  ");
      if(idx->flags & ANY_SRC_PORT)
         printf("ANY_SRC_PORT  ");
      if(idx->flags & ANY_DST_IP)
         printf("ANY_DST_IP  ");
      if(idx->flags & ANY_DST_PORT)
         printf("ANY_DST_PORT");
      printf("\n");

      if(idx->message != NULL)
      {
         printf("      message: %s\n", idx->message);
      }

      idx = idx->next;
   }
}
 


